/**sfh
******************************************************************************

	stream
	~~~~~~
		 
		Implements the stream fetch mode where tiles are download as a 
		continuous stream, and state/event data is provided in cookies.

******************************************************************************
efh*/

//
// Objects
//
var Debug = GetDebugObject();
var CookieMgrFactory = CreateCookieMgrObject;
var Engine = GetEngineObject();
var Comms = GetCommsObject();
var LocalAPI = GetLocalAPIObject();
var StreamMode = GetStreamModeObject();

function GetStreamModeObject()
{
	if (!GetStreamModeObject.StreamMode)
	{
		GetStreamModeObject.StreamMode = new StreamModeObject;
	}
	return GetStreamModeObject.StreamMode;
}

/*sfh
******************************************************************************

	Function:		StreamModeObject

	Desc:			Object which manages tile fetching while in STREAM fetch
					mode.

	Returns:		void

******************************************************************************
efh*/

function StreamModeObject()
{
	//
	// Statics
	// 
	var s_FETCH_MODE_CACHE = 0;				// Poll for a list of dirty tiles using 'getupdate'; supports caching
	var s_FETCH_MODE_STREAM = 1;			// Download next tile directly to image object using 'getnexttile'

	var s_DEFAULT_TILE_CACHE_SIZE = 10;		// Default size of tile cache used when in STREAM fetch mode in lieu of browser cache
	var s_DEFAULT_CURSOR_CACHE_SIZE = 10;	// Default size of cursor cache used when in STREAM fetch mode in lieu of browser cache
	var s_GETNEXTTILE_TIMEOUT = 15000;  	// Number of milliseconds we will wait for a 'getnexttile' request before retry
	var s_GETNEXTTILE_WAITSTATE = 1;			// millis - wait time between calls to 'getnexttile' - keep very low

	//
	// Vars
	//
	var m_fetchMode = s_FETCH_MODE_STREAM;				// Current fetch mode
	var m_iTileCacheSize = s_DEFAULT_TILE_CACHE_SIZE;	// Cache sizes used for stream fetch mode
	var m_iCursorCacheSize = s_DEFAULT_CURSOR_CACHE_SIZE;	
	var m_arrTileCacheImg = new Object();				// Tile cache storing the img and indexed by crc 
	var m_arrTileCacheCrc = new Array();				// Tile cache storing the crc and indexed by ordinal
	var m_arrCursorCacheImg = new Object();				// Cursor cache storing the img and indexed by crc
	var m_arrCursorCacheCrc = new Array();				// Cursor cache storing the crc and indexed by ordinal
	var m_bResetTileCache = true;						// When true will tell server to reset its tile cache
	var m_bResetCursorCache = true;						// When true will tell server to reset its cursor cache
	var m_seq = 0;										// Sequence number for each request

	// Session
	var m_assid = null;
	var m_cid = null;
	var m_ck = null; 
	
	//
	// Access methods
	//
	this.GetDefaultTileCacheSize = function() { return s_DEFAULT_TILE_CACHE_SIZE; };
	this.GetDefaultCursorCacheSize = function() { return s_DEFAULT_CURSOR_CACHE_SIZE; };
	this.SetTileCacheSize = function(size) { m_iTileCacheSize = size; };
	this.SetCursorCacheSize = function(size) { m_iCursorCacheSize = size; };
	this.GetFetchMode = function() { return m_fetchMode; };
	this.SetSessionInfo = function(assid, cid, ck) { m_assid=assid; m_cid=cid; m_ck=ck; };


	/*sfh
	******************************************************************************
	
		Function:		SetFetchMode
	
		Desc:			Sets the current fetch mode. This does not really have 
						any effect, but is here to record which fetch mode
						the server confirmed. Aegis automatically works in either
						mode so no action is required here.
	
		Returns:		void
	
	******************************************************************************
	efh*/
	
	this.SetFetchMode = function(fetchMode)
	{
		m_fetchMode = fetchMode;
	};

	/*sfh
	******************************************************************************
	
		Function:		IsValidFetchMode
	
		Desc:			Returns true if the specified fetch mode value is valid
						as per the constants defined.
	
		Returns:		boolean
	
	******************************************************************************
	efh*/
	
	this.IsValidFetchMode = function(fetchMode)
	{
		return (fetchMode == s_FETCH_MODE_CACHE || fetchMode == s_FETCH_MODE_STREAM) ? true : false;
	};

	/**sfh
	******************************************************************************
	
		Function:		GetDefaultFetchMode
	
		Desc:			Returns the default fetch mode for this browser.
						IE6 always defaults to CACHE mode because of bugs in
						image handling and caching. Otherwise IE6 would not
						benefit from STREAM mode's caching feature and would
						dwonload tiles multiple times.<b> 
						
		Returns:		void
	
	******************************************************************************
	efh*/
	
	this.GetDefaultFetchMode = function()
	{
		// Get the version number of IE.
		if (navigator.appName == 'Microsoft Internet Explorer')
		{
			var ua = navigator.userAgent;
			var re  = new RegExp("MSIE ([0-9]{1,}[\.0-9]{0,})");
			var ver = 0;
			if (re.exec(ua) !== null)
			{
				ver = parseFloat( RegExp.$1 );
			}
		}
		return (ver == 6) ? s_FETCH_MODE_CACHE : s_FETCH_MODE_STREAM;
	};

	/*sfh
	******************************************************************************
	
		Function:		GetNextTile
	
		Desc:			When in fetch mode s_FETCH_MODE_STREAM this method
						will fetch the next tile via an image object, and continue
						fetching tiles until none are left.
						
						When no tiles are left to fetch, then the callback is 
						called and the 'getnexttile' loop ends.
	
		Returns:		void
	
	******************************************************************************
	efh*/
	
	this.GetNextTile = function(callback)
	{
		// Clean up last request.
		if (this.GetNextTile.timer)
		{
			window.clearTimeout(this.GetNextTile.timer);
			this.GetNextTile.timer = 0;
		}

		// Remember the callback.
		if (typeof callback != "undefined")
		{
			this.GetNextTile.callback = callback;
		}

		// Don't create the image callbacks every time.
		if (!this.GetNextTile.onloadHandler)
		{
			this.GetNextTile.onLoadHandler = new Function("StreamMode.On_GetNextTile_Loaded();");
			this.GetNextTile.onErrorHandler = new Function("StreamMode.On_GetNextTile_Error('error');");
			this.GetNextTile.onAbortHandler = new Function("StreamMode.On_GetNextTile_Error('abort');");
		}

		//  Create new image object to receive the requested tile (or cursor).
		this.GetNextTile.img = document.createElement("img");
		this.GetNextTile.img.onload = this.GetNextTile.onLoadHandler;
		this.GetNextTile.img.onerror = this.GetNextTile.onErrorHandler;
		this.GetNextTile.img.onabort = this.GetNextTile.onAbortHandler;

		// Set up a timeout timer for the request
		this.GetNextTile.timer = window.setTimeout("StreamMode.On_GetNextTile_Error('timeout');", s_GETNEXTTILE_TIMEOUT);

		// Handle cache reset flags.
		var strReset = "";
		if (m_bResetTileCache) 
		{ 
			strReset = strReset + "/rct=1"; 
			m_arrTileCacheImg = new Object();
			m_arrTileCacheCrc = new Array();
		}
		if (m_bResetCursorCache) 
		{ 
			strReset = strReset + "/rcc=1"; 
			m_arrCursorCacheImg = new Object();
			m_arrCursorCacheCrc = new Array(); 
		}
		m_bResetTileCache = false;
		m_bResetCursorCache = false;

		// Start the download.
		var url = Comms.URL_GET_NEXT_TILE + "assid=" + m_assid + "/cid=" + m_cid + "/ck=" + m_ck + strReset;
		url = url + "/seq=" + m_seq + "/ts=" + (new Date().getTime());
		//Debug.ConsoleOut("GetNextTile:  requesting url... " + url);
		this.GetNextTile.img.src = url;
	};	

	/*sfh
	******************************************************************************
	
		Function:		On_GetNextTile_Loaded
	
		Desc:			Called when in fetch mode s_FETCH_MODE_STREAM and a
						'getnexttile' image object request is completed.
	
		Returns:		void
	
	******************************************************************************
	efh*/
	
	this.On_GetNextTile_Loaded = function()
	{
		// Kill the timeout.
		if (this.GetNextTile.timer)
		{
			window.clearTimeout(this.GetNextTile.timer);
			this.GetNextTile.timer = 0;
		}
	
		// Create a cookie mgr for parsing the cookies returned.
		var cookieMgr_Msg = CookieMgrFactory(document.cookie);
		var cookieMgr_Payload = CookieMgrFactory(cookieMgr_Msg.GetCookieWithDefault("msg"), null);

		// First check for the nodata directive.
		var bNoData = cookieMgr_Payload.GetCookieWithDefault("nodata", 0);

		Debug.ConsoleOut("Dump msg cookies...  " + cookieMgr_Payload.ToString());

		if (!bNoData)
		{
			try
			{
				// Let the LocalAPI object process any commands provided via the server generated cookie.
				LocalAPI.ProcessCookies(cookieMgr_Payload);
		
				// Process any tile image update.
				ProcessTileUpdate(this.GetNextTile.img, cookieMgr_Payload);
				ProcessCursorUpdate(this.GetNextTile.img, cookieMgr_Payload);
			}
			catch(e)
			{
				Debug.ConsoleOut("!!! Exception caught in On_GetNextTile_Loaded");
				Debug.ConsoleOutException(e);
			}

			// Loop to next request - wait a few milliseconds then start another 'getnexttile'.
		 	window.setTimeout("StreamMode.GetNextTile();", s_GETNEXTTILE_WAITSTATE);
		}
		else
		{
			// We're done with the 'getnexttile' loop for now. Call the callback then return.
			if (this.GetNextTile.callback) { this.GetNextTile.callback(); }
		}

		// If no errors, then we can update the sequence number to indicate to the server that 
		// this request was completed, and we won't need a re-send.
		m_seq = cookieMgr_Payload.GetCookie("seq");

		// Clear all cookie values for the next request, but preserve the client id.
		var cid = cookieMgr_Msg.GetCookie("cid");
		cookieMgr_Msg.DeleteAllCookies();
		cookieMgr_Msg.SetSessionCookie("cid", cid);
	};

	/*sfh
	******************************************************************************
	
		Function:		On_GetNextTile_Error
	
		Desc:			Called when in fetch mode s_FETCH_MODE_STREAM and a
						'getnexttile' image object request ended in error or
						timed out.
	
		Returns:		void
	
	******************************************************************************
	efh*/
	
	this.On_GetNextTile_Error = function(strErrorType)
	{
		// Kill the timeout.
		if (this.GetNextTile.timer)
		{
			if (strErrorType != "timeout") { window.clearTimeout(this.GetNextTile.timer); }
			this.GetNextTile.timer = 0;
		}

		Debug.ConsoleOut("On_GetNextTile_Error - errortype=" + strErrorType);

		// Since there was an error we need to exit the 'getnexttile' loop for now. Call the callback then return.
		if (this.GetNextTile.callback) { this.GetNextTile.callback(); }
	};

	/*sfh
	******************************************************************************
	
		Function:		ProcessTileUpdate
	
		Desc:			Called when a new tile has been received by 'getnexttile'.
						Here we check if the tile is already in our cache, and 
						handle background tiles.
	
		Returns:		bool - true if successful, false if request should be 
		                       repeated.
	
	******************************************************************************
	efh*/

	function ProcessTileUpdate(img, cookieMgr)
	{
		var strTids = cookieMgr.GetCookieWithDefault("tid", null);
		if (strTids === null)  { return true; }

		var bError = false;

		// Process any tile image updates (there may be more than one destination tile for the image).
		var arrTids = strTids.split(",");
		var bBKTile = (arrTids !== null && arrTids.length == 1 && arrTids[0] == "0") ? true : false; 
		var bNegTileID = (arrTids !== null && arrTids.length > 0 && parseInt(arrTids[0]) < 0) ? true : false;
		var bCached = cookieMgr.GetCookieBool("cache");
		var crc = cookieMgr.GetCookieWithDefault("crc", null);
		var bid = cookieMgr.GetCookieIntWithDefault("bid", 0);
		var bLastInBatch = cookieMgr.GetCookieIntWithDefault("bidl", 0);

		if (bBKTile)
		{
			Engine.EngMarkTileDirty(0, bid, img.src, bLastInBatch);
			Debug.ConsoleOut("ProcessTileUpdate -- background tile ---  crc=" + crc + "; imgSize=" + img.width + "x" + img.height);		
		}
		else if (bCached)
		{
			// Find the crc in our cache.
			if (crc !== null && m_arrTileCacheImg[crc])
			{
				// Cache hit! Use the cached image as the image downloaded is blank.
				Debug.ConsoleOut("ProcessTileUpdate -- cache HIT ---  crc=" + crc + "; imgSize=" + img.width + "x" + img.height + "; tids=" + strTids);
				Debug.ConsoleOutProps(m_arrTileCacheCrc);
				img = m_arrTileCacheImg[crc];
			}
			else
			{
				// Cache error! This means the Aegis cache is out of sync with the server cache.
				// Reset the server cache and return error so this request is repeated.
				bError = true;
				m_bResetTileCache = true;
				Debug.ConsoleOut("ProcessTileUpdate -- cache ERROR ---  crc=" + crc + "; tids=" + strTids);
			}
		}
		else if (!bNegTileID && !bCached)
		{
			// Add this image to the cache (if not already there).
			if (crc !== null && !m_arrTileCacheImg[crc])
			{
				// Honour the maximum cache size by killing the oldest as needed.
				//    Array layout is : m_arrTileCacheImg['crcxxx'] = img
				//	  					m_arrTileCacheCrc[0] = 'crcxxx'
				if (m_arrTileCacheCrc.length >= m_iTileCacheSize)
				{
					var delCrc = m_arrTileCacheCrc[0];
					m_arrTileCacheCrc.shift(); 			// Deletes element at index 0
					delete m_arrTileCacheImg[delCrc];	// Deletes the property
				}
				m_arrTileCacheCrc.push(crc);	// Add to end of array
				m_arrTileCacheImg[crc] = img;	// Create property with crc as the property name

				//Debug.ConsoleOut("ProcessTileUpdate -- cache MISS --- image added --- crc=" + crc + "; imgSize=" + img.width + "x" + img.height + "; tids=" + strTids);
			}
			else
			{
				// The server thinks this is not in our cache, but it is!
				Debug.ConsoleOut("ProcessTileUpdate -- cache MISS --- image already in cache!!! --- crc=" + crc + "; tids=" + strTids);
			}
		}

		// Assign this image to all the tiles listed.
		if (!bError && !bBKTile)
		{
			var iTileID;
			for (var i = 0;  i < arrTids.length;  i = i + 1)
			{
				try
				{
					iTileID = parseInt(arrTids[i]);
					
					// Force the engine to update this tile. Since the image
					// is already downloaded it will be pulled from the browser cache.
					Engine.EngMarkTileDirty(iTileID, bid, img.src, bLastInBatch);
				}
				catch(e)
				{
					Debug.ConsoleOut("!!!  ProcessTileUpdate -- error updating tile: " + arrTids[i]);
					Debug.ConsoleOutException(e);
				}
			}
		}
		return !bError;
	}

	/*sfh
	******************************************************************************
	
		Function:		ProcessCursorUpdate
	
		Desc:			Called when a new cursor has been received by 'getnexttile'.
						Here we check if the cursor is already in the cache.
	
		Returns:		bool - true if successful, false if request should be 
		                       repeated.
	
	******************************************************************************
	efh*/

	function ProcessCursorUpdate(img, cookieMgr)
	{
		if (!cookieMgr.GetCookieBool("micon"))  { return true; }
	
		var bError = false;
		var bCached = cookieMgr.GetCookieBool("cache");
		var crc = cookieMgr.GetCookieWithDefault("crc", null);

		if (bCached)
		{
			// Find the crc in our cache.
			if (crc !== null && m_arrCursorCacheImg[crc])
			{
				// Cache hit! Use the cached image as the image downloaded is blank.
				img = m_arrCursorCacheImg[crc];
				Debug.ConsoleOut("ProcessCursorUpdate -- cache HIT ---  crc=" + crc + "; imgSize=" + img.width + "x" + img.height);
			}
			else
			{
				// Cache error! This means the Aegis cache is out of sync with the server cache.
				// Reset the server cache and return error so this request is repeated.
				bError = true;
				m_bResetCursorCache = true;
				Debug.ConsoleOut("ProcessCursorUpdate -- cache ERROR ---  crc=" + crc);
			}
		}
		else // !bCached
		{
			// Add this image to the cache (if not already there).
			if (crc !== null && !m_arrCursorCacheImg[crc])
			{
				// Honour the maximum cache size by killing the oldest as needed.
				//    Array layout is : m_arrCursorCacheImg['crcxxx'] = img
				//	  					m_arrCursorCacheCrc[0] = 'crcxxx'
				if (m_arrCursorCacheCrc.length >= m_iCursorCacheSize)
				{
					var delCrc = m_arrCursorCacheCrc[0];
					m_arrCursorCacheCrc.shift(); 			// Deletes element at index 0
					delete m_arrCursorCacheImg[delCrc];		// Deletes the property
				}
				m_arrCursorCacheCrc.push(crc);	// Add to end of array
				m_arrCursorCacheImg[crc] = img;	// Create property with crc as the property name

				Debug.ConsoleOut("ProcessCursorUpdate -- cache MISS --- image added --- crc=" + crc + "; imgSize=" + img.width + "x" + img.height);
			}
			else
			{
				Debug.ConsoleOut("ProcessCursorUpdate -- cache MISS --- image already in cache!!! --- crc=" + crc);
			}
		}

		if (!bError)
		{
			// Set the cursor image.
			Engine.EngMarkCursorImageDirty(img.src);
		}
		return !bError;
	}

} // End of StreamModeObject 
